###################################################################################################
# _metaobjmanager.py
#
# Implementation of class MetaobjManager (see below).
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
###################################################################################################

# Imports.
from Globals import HTML, HTMLFile
import copy
import string
import urllib
# Product Imports.
import _globals
import _xmllib


# -------------------------------------------------------------------------------------------------
#  importMetaObj
# -------------------------------------------------------------------------------------------------

def _importMetaObj(self, item):
  key = item['key']
  value = item['value']
  dtml = item['dtml']
  self.__init_metaobj__(key,value,dtml)
  self.recurse_updateVersionMetaObjs()
  return key

def importMetaObj(self, xml):
  builder = _xmllib.XmlAttrBuilder()
  value = builder.parse(xml)
  if type(value) is type([]):
    for item in value:
      id = _importMetaObj(self,item)
  else:
    id = _importMetaObj(self,value)
  self.synchronizeObjAttrs()
  return id


# -------------------------------------------------------------------------------------------------
#  findMetaObj
# -------------------------------------------------------------------------------------------------
def findMetaObj(self, meta_id):
  if self.meta_type == 'ZMSCustom' and self.meta_id == meta_id:
    return self
  for child in self.objectValues():
    metaobj = findMetaObj(child,meta_id)
    if metaobj is not None:
      return metaobj
  return None

# -------------------------------------------------------------------------------------------------
#  getTemplateId
# -------------------------------------------------------------------------------------------------
def getTemplateId(id):
  return "bodyContentZMSCustom_%s"%id


###################################################################################################
###################################################################################################
###
###   class MetaobjManager
###
###################################################################################################
###################################################################################################
class MetaobjManager: 


    # Management Interface.
    # ---------------------
    manage_customizeMetaobjForm = HTMLFile('dtml/metaobj/manage_customizeform', globals()) 
    metaobj_input_fields = HTMLFile('dtml/metaobj/input_fields', globals())
    metaobj_input_js = HTMLFile('dtml/metaobj/input_js', globals())


    # ------------------------------------------------------------------------------------------
    # MetaobjManager.recurse_updateVersionMetaObjs: 
    #
    # Update version.
    # ------------------------------------------------------------------------------------------
    def recurse_updateVersionMetaObjs(self):
      #++ if _globals.debug: print "[%s.recurse_updateVersionMetaObjs]"%self.meta_type
      message = ""

      # -----------------------------------------------------------
      # 17.04.2002
      # Build #071: Meta-Object Type.
      # -----------------------------------------------------------
      if getattr(self.getDocumentElement(),'build','') < '071':
        obs = self.__get_metaobjs__()
        for id in obs.keys():
          ob = self.__get_metaobj__(id)
	  tmplt_ids = self.getHome().objectIds(['DTML Method'])
	  if 'ZMSCustom_%s'%id in tmplt_ids:
	    ob['type'] = 'ZMSObject'
	  else:
	    ob['type'] = 'ZMSDocument'
          obs[id] = ob
        self.setConfProperty('ZMS.custom.objects',obs.copy())

      # -----------------------------------------------------------
      # 20.08.2002
      # Build #094: Meta-Object Attribute Keys.
      # -----------------------------------------------------------
      if getattr(self.getDocumentElement(),'build','') < '094':
        obs = self.__get_metaobjs__()
        for id in obs.keys():
          ob = self.__get_metaobj__(id)
          attrs = []
          for attr in ob['__obj_attrs__']:
            attr_keys = ''
            if attr.has_key('keys'):
              attr_keys = attr['keys']
              if type(attr_keys) is type(''):
                attr_keys = string.replace(attr_keys,',','\n')
            attr['keys'] = attr_keys
            attrs.append(attr)
          ob['__obj_attrs__'] = attrs
          obs[id] = ob
        self.setConfProperty('ZMS.custom.objects',obs.copy())

      # -----------------------------------------------------------
      # 22.08.2002
      # Build #096: Replace Meta-Object Attribute Type Html by Text.
      # -----------------------------------------------------------
      if getattr(self.getDocumentElement(),'build','') < '096':
        obs = self.__get_metaobjs__()
        for id in obs.keys():
          ob = self.__get_metaobj__(id)
          attrs = []
          for attr in ob['__obj_attrs__']:
            attr_type = attr['type']
            if attr_type == 'LangHtml':
              attr_type = 'LangText'
            attr['type'] = attr_type
            attrs.append(attr)
          ob['__obj_attrs__'] = attrs
          obs[id] = ob
        self.setConfProperty('ZMS.custom.objects',obs.copy())
      
      # -----------------------------------------------------------
      # 16.12.2002
      # Build #108: Move Templates from Home- to Document-Element.
      # -----------------------------------------------------------
      if getattr(self.getDocumentElement(),'build','') < '108':
        obs = self.__get_metaobjs__()
        for obId in obs.keys():
          ob = self.__get_metaobj__(obId)
          obType = ob['type']
          if obType == 'ZMSDocument':
            id = "bodyContentZMSCustom_%s"%obId
          elif obType == 'ZMSTeaserElement':
            id = "bodyContentTeaserZMSCustom_%s"%obId
          else:
            id = "ZMSCustom_%s"%obId
          new_id = getTemplateId(obId)
          homeElmnt = self.getHome()
          docElmnt = self.getDocumentElement()
          try:
            cb_copy_data = homeElmnt.manage_cutObjects(ids=[id])
            docElmnt.manage_pasteObjects(cb_copy_data)
            if 'copy_of_%s'%id in docElmnt.objectIds():
              id = 'copy_of_%s'%id
            docElmnt.manage_renameObject(id=id,new_id=new_id)
          except:
            pass

      # -----------------------------------------------------------
      # 19.12.2002
      # Build #109: Allow non-multilang attributes.
      # -----------------------------------------------------------
      obs = self.__get_metaobjs__().copy()
      for id in obs.keys():
        ob = obs[id]
        attrs = []
        for attr in ob['__obj_attrs__']:
          is_objecttype = attr['type'] in self.dGlobalAttrs.keys() or attr['type'] in self.getMetaobjIds()
          attr['multilang'] = attr.get('multilang',not is_objecttype or attr['type'][:4]=='Lang')
          if type(attr['keys']) is type(''):
            attr['keys'] = self.string_list(attr['keys'],'\n')
          if not is_objecttype and attr['type'][:4]=='Lang':
            attr_keys = attr['keys']
            attr_type = string.lower(attr['type'][4:])
            if len(attr_keys) > 0:
              if attr_type == 'string':
                attr_type = 'select'
              elif attr_type == 'text':
                attr_type = 'multiselect'
              else:
                attr_keys = []
            attr['keys'] = attr_keys
            attr['type'] = attr_type
          attrs.append(attr)
        ob['__obj_attrs__'] = attrs
        obs[id] = ob
      self.setConfProperty('ZMS.custom.objects',obs.copy())

      return message

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.__get_metaobjs__:
    # 
    # Returns dictionary of Meta-Objects.
    # ------------------------------------------------------------------------------------------
    def __get_metaobjs__(self):
      return self.getConfProperty('ZMS.custom.objects',{})

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.__get_metaobj__:
    # 
    # Returns Meta-Object identified by ID.
    # ------------------------------------------------------------------------------------------
    def __get_metaobj__(self, id): 
      return self.__get_metaobjs__()[id]

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.getMetaobjIds:
    # 
    # Returns list of IDs of Meta-Objects.
    # ------------------------------------------------------------------------------------------
    def getMetaobjIds(self, sort=1):
      metaobjs = self.__get_metaobjs__()
      ids = metaobjs.keys()
      if sort:
        mapping = []
        for id in ids:
          ob = metaobjs[id]
          mapping.append((ob['name'],id))
        mapping.sort()
        ids = map(lambda x: x[1],mapping)
      return ids

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.getMetaobj:
    # 
    # Returns Meta-Object identified by ID.
    # ------------------------------------------------------------------------------------------
    def getMetaobj(self, id):
      ob = self.__get_metaobj__(id)
      # Retrieve properties.
      name = ob['name']
      type = ob['type']
      attrs = ob['__obj_attrs__']
      tmplt = getTemplateId(id)
      # Return dictionary of properties.
      return {'id':id,'name':name,'type':type,'attrs':attrs,'tmplt':tmplt}

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.getMetaobjId:
    # 
    # Returns Meta-Object ID identified by Name.
    # ------------------------------------------------------------------------------------------
    def getMetaobjId(self, name):
      for id in self.getMetaobjIds():
        ob = self.getMetaobj(id)
        if name == ob['name']:
          return id
      return None


    """
    ###############################################################################################
    #
    #   A T T R I B U T E S
    #
    ###############################################################################################
    """

    # ------------------------------------------------------------------------------------------
    # MetaobjManager._initMetaObjAttrs:
    # 
    # Init list of Attributes for Meta-Object specified by ID.
    # ------------------------------------------------------------------------------------------
    def _initMetaObjAttrs(self, id):
      attrs = []
      cp = copy.copy(attrs)
      # Init Attributes.
      ob = self.__get_metaobj__(id)
      ob['__obj_attrs__'] = attrs
      # Set Attributes to Meta-Object.
      obs =  self.__get_metaobjs__()
      obs[id] = ob
      self.setConfProperty('ZMS.custom.objects',obs.copy())

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.getMetaobjAttrIds:
    # 
    # Returns list of Attribute-IDs for Meta-Object specified by ID.
    # ------------------------------------------------------------------------------------------
    def getMetaobjAttrIds(self, meta_id):
      ids = []
      ob = self.__get_metaobj__(meta_id)
      attrs = ob['__obj_attrs__']
      for attr in attrs:
        ids.append(attr['id'])
      return ids

    # ------------------------------------------------------------------------------------------
    # MetaobjManager.getMetaobjAttr:
    # 
    # Get Attribute for Meta-Object specified by key.
    # ------------------------------------------------------------------------------------------
    def getMetaobjAttr(self, meta_id, key):
      ob = self.__get_metaobj__(meta_id)
      attrs = ob['__obj_attrs__']
      for attr in attrs:
        if attr['id'] == key:
          attr = attr.copy()
          attr['mandatory'] = attr.get('mandatory',0)
          attr['multilang'] = attr.get('multilang',1)
          attr['size'] = attr.get('size',35)
          return attr
      return None

    # ------------------------------------------------------------------------------------------
    # MetaobjManager._setMetaObjAttr:
    # 
    # Set/add Meta-Object Attribute with specified values.
    # ------------------------------------------------------------------------------------------
    def _setMetaObjAttr(self, id, attr_id, newName, newMandatory=0, newMultilang=1, newRepetitive=0, newType='string', newKeys=[]):
      ob = self.__get_metaobj__(id)
      attrs = copy.copy(ob['__obj_attrs__'])
      # Set Attributes.
      if not newType in ['*','multiselect','select']:
        newKeys = []
      if newType=='*' or newType in self.dGlobalAttrs.keys() or newType in self.getMetaobjIds():
        newMultilang = 0
      else:
        newRepetitive = 0
        
      attr = {}
      attr['id'] = attr_id
      attr['name'] = newName
      attr['mandatory'] = newMandatory
      attr['multilang'] = newMultilang
      attr['repetitive'] = newRepetitive
      attr['type'] = newType
      attr['keys'] = newKeys
      
      ids = self.getMetaobjAttrIds(id)
      if attr_id in ids:
        attrs[ids.index(attr_id)] = attr
      else:
        attrs.append(attr)
      ob['__obj_attrs__'] = attrs
      # Set Attributes to Meta-Object.
      obs =  self.__get_metaobjs__()
      obs[id] = ob
      self.setConfProperty('ZMS.custom.objects',obs.copy())

    # ------------------------------------------------------------------------------------------
    # MetaobjManager._moveMetaObjAttr:
    # 
    # Moves Meta-Object Attribute in specified direction (-1=up, +1=down).
    # ------------------------------------------------------------------------------------------
    def _moveMetaObjAttr(self, id, attr_id, direction=0):
      ob = self.__get_metaobj__(id)
      attrs = copy.copy(ob['__obj_attrs__'])
      # Move Attributes.
      ids = self.getMetaobjAttrIds(id)
      i = ids.index(attr_id)
      attr = attrs[i]
      attrs[i] = attrs[i+direction]
      attrs[i+direction] = attr
      ob['__obj_attrs__'] = attrs
      # Set Attributes to Meta-Object.
      obs =  self.__get_metaobjs__()
      obs[id] = ob
      self.setConfProperty('ZMS.custom.objects',obs.copy())
      
    # ------------------------------------------------------------------------------------------
    # MetaobjManager._delMetaObjAttr:
    # 
    # Delete Attribute from Meta-Object specified by ID..
    # ------------------------------------------------------------------------------------------
    def _delMetaObjAttr(self, id, attr_id):
      # Delete Attribute.
      ob = self.__get_metaobj__(id)
      cp = []
      attrs = copy.copy(ob['__obj_attrs__'])
      for attr in attrs:
        if attr['id'] != attr_id:
          cp.append(attr)
      ob['__obj_attrs__'] = cp
      # Set Attributes to Meta-Object.
      obs =  self.__get_metaobjs__()
      obs[id] = ob
      self.setConfProperty('ZMS.custom.objects',obs.copy())

    ###############################################################################################

    # ------------------------------------------------------------------------------------------
    #  MetaobjManager.__init_metaobj__:
    # 
    #  Init Meta-Object with specified values.
    # ------------------------------------------------------------------------------------------
    def __init_metaobj__(self, id, ob, dtml_raw=None):
    
      obs =  self.__get_metaobjs__()
      
      # Template.
      # ---------
      docElmnt = self.getDocumentElement()
      dtml_id = getTemplateId(id)
      dtml_title = "Template: %s"%ob['name']
      if dtml_raw is None:
        dtml_raw = ""
        dtml_raw = dtml_raw + "<!-- BO %s -->\n"%dtml_id
        dtml_raw = dtml_raw + "<!-- EO %s -->\n"%dtml_id
      # Insert.
      if not hasattr(docElmnt,dtml_id):
        docElmnt.manage_addDTMLMethod(dtml_id,dtml_title,dtml_raw)
      # Proxy.
      dtml_tmplt = getattr(docElmnt,dtml_id)
      dtml_tmplt.manage_proxy(roles=['Manager'])
      
      # Set Attributes to Meta-Object.
      obs[id] = ob
      self.setConfProperty('ZMS.custom.objects',obs.copy())


    # ------------------------------------------------------------------------------------------
    # MetaobjManager.__set_metaobj__:
    # 
    # Set/add Meta-Object with specified values.
    # ------------------------------------------------------------------------------------------
    def __set_metaobj__(self, id, name, metatype, attrs=[]):
      
      # Init Meta-Object.
      # -----------------
      obs = self.__get_metaobjs__()
      
      # Set Meta-Object.
      # ----------------
      ob = {}
      ob['name'] = name
      ob['type'] = metatype
      ob['__obj_attrs__'] = copy.copy(attrs)
      self.__init_metaobj__(id,ob)


    # ------------------------------------------------------------------------------------------
    # MetaobjManager.__del_metaobj__:
    # 
    # Delete Meta-Object specified by ID.
    # ------------------------------------------------------------------------------------------
    def __del_metaobj__(self, id):
      
      # Delete from attribute values.
      cp = self.__get_metaobjs__()
      obs = {}
      for key in cp.keys():
        if key != id:
          obs[key] = cp[key]
      
      # Delete Template.
      docElmnt = self.getDocumentElement()
      dtml_id = getTemplateId(id)
      if hasattr(docElmnt,dtml_id):
        docElmnt.manage_delObjects(ids=[dtml_id])
      
      # Set Attributes to Meta-Object.
      self.setConfProperty('ZMS.custom.objects',obs.copy())


    ###############################################################################################
    # MetaobjManager.manage_changeMetaobjs:
    #
    # Werte fr Meta-Objects lschen oder ndern.
    ###############################################################################################
    def manage_changeMetaobjs(self, lang, manage_lang, btn='', key='all', REQUEST=None, RESPONSE=None):
        """ MetaobjManager.manage_changeMetaobjs """
        
        message = ''
        id = REQUEST.get('id','')
        
        # Delete.
        # -------
        
        # Obj.
        if key == 'obj' and btn == self.getLangStr('BTN_DELETE',manage_lang):
          metaobj = findMetaObj(self,id)
          if metaobj is None:
            self.__del_metaobj__(id)
            id = ''
            message = self.getLangStr('MSG_CHANGED',manage_lang)
          else:
            message = '[ZMS-ERROR]: Constraint violation at <a href="%s/manage_main?lang=%s&manage_lang=%s#_%s">%s</a>'%(metaobj.getParentNode().absolute_url(),lang,manage_lang,metaobj.id,metaobj.absolute_url())
        
        # Attr.
        elif key == 'attr' and btn == 'delete':
          attr_id = REQUEST['attr_id']
          self._delMetaObjAttr(id,attr_id)
        
        # Change.
        # -------
        elif key == 'all' and btn == self.getLangStr('BTN_CHANGE',manage_lang):
        
          # Obj.
          id = string.strip(REQUEST['id'])
          name = string.strip(REQUEST['obj_name'])
          metatype = REQUEST['obj_type']
          attrs = ob = self.__get_metaobj__(id)
          attrs = ob['__obj_attrs__']
          self.__set_metaobj__(id,name,metatype,attrs)
          metaObj = self.getMetaobj(id)
          if hasattr(self,metaObj['tmplt']):
            ob = getattr(self,metaObj['tmplt'])
            ob.manage_edit(title='Template: %s'%metaObj['name'],data=REQUEST.get('obj_dtml',''))
            ob.manage_proxy(roles=['Manager'])
          
          # Attr.
          saved_ids = self.getMetaobjAttrIds(id)
          self._initMetaObjAttrs(id)
          for saved_id in saved_ids:
            attr_id = string.strip(REQUEST['attr_id_%s'%saved_id])
            newName = string.strip(REQUEST['attr_name_%s'%saved_id])
            newMandatory = REQUEST.get('attr_mandatory_%s'%saved_id,0)
            newMultilang = REQUEST.get('attr_multilang_%s'%saved_id,0)
            newRepetitive = REQUEST.get('attr_repetitive_%s'%saved_id,0)
            newType = REQUEST['attr_type_%s'%saved_id]
            newKeys = self.string_list(REQUEST.get('attr_keys_%s'%saved_id,''),'\n')
            self._setMetaObjAttr(id,attr_id,newName,newMandatory,newMultilang,newRepetitive,newType,newKeys)
          
          message = self.getLangStr('MSG_CHANGED',manage_lang)
        
        # Export.
        # -------
        elif btn == self.getLangStr('BTN_EXPORT',manage_lang):
          value = []
          ids = REQUEST.get('ids',[])
          for id in self.__get_metaobjs__().keys():
            if id in ids or len(ids) == 0:
              ob = self.__get_metaobj__(id)
              docElmnt = self.getDocumentElement()
              dtml_id = getTemplateId(id)
              dtml_raw = ""
              for method in docElmnt.objectValues(['DTML Method']):
                if method.id() == dtml_id:
                  dtml_raw = method.raw
              value.append({'key':id,'value':ob,'dtml':dtml_raw})
          if len(value)==1:
            value = value[0]
          content_type = 'text/xml'
          filename = id + '.xml'
          export = '<?xml version="1.0" encoding="ISO-8859-1"?>'
          export = export + _xmllib.toXml(self,value,1)
          RESPONSE.setHeader('Content-Type',content_type)
          RESPONSE.setHeader('Content-Disposition','inline;filename=%s'%filename)
          return export
        
        # Insert.
        # -------
        elif btn == self.getLangStr('BTN_INSERT',manage_lang):
        
          # Obj.
          if key == 'obj':
            id = string.strip(REQUEST['_meta_id'])
            name = string.strip(REQUEST['_meta_name'])
            metatype = string.strip(REQUEST['_meta_type'])
            self.__set_metaobj__(id,name,metatype)
            if metatype == 'ZMSDocument':
              self._setMetaObjAttr(id,'titlealt',self.getLangStr('ATTR_TITLESHORT',manage_lang),1,1,0,'string')
              self._setMetaObjAttr(id,'title',self.getLangStr('ATTR_TITLE',manage_lang),1,1,0,'string')
            elif metatype == 'ZMSTeaserElement':
              self._setMetaObjAttr(id,'titlealt',self.getLangStr('ATTR_TITLESHORT',manage_lang),1,1,0,'string')
              self._setMetaObjAttr(id,'attr_penetrance',self.getLangStr('ATTR_PENETRANCE',manage_lang),1,1,0,'select',['this','sub_nav','sub_all'])
          
          # Attr.
          if key == 'attr':
            attr_id = string.strip(REQUEST['attr_id'])
            newName = string.strip(REQUEST['attr_name'])
            newMandatory = REQUEST.get('_mandatory',0)
            newMultilang = REQUEST.get('_multilang',0)
            newRepetitive = REQUEST.get('_repetitive',0)
            newType = REQUEST['attr_type']
            newKeys = self.string_list(REQUEST.get('_keys',''),'\n')
            self._setMetaObjAttr(id,attr_id,newName,newMandatory,newMultilang,newRepetitive,newType,newKeys)
        
        # Import.
        # -------
        elif btn == self.getLangStr('BTN_IMPORT',manage_lang):
          id = importMetaObj(self,xml=REQUEST['_meta_file'])
          message = self.getLangStr('MSG_IMPORTED',manage_lang)%('<i>%s</i>'%REQUEST['_meta_file'].filename)

        # Move up.
        # --------
        elif key == 'attr' and btn == 'move_up':
          attr_id = REQUEST['attr_id']
          self._moveMetaObjAttr(id,attr_id,-1)
          message = self.getLangStr('MSG_MOVEDOBJUP',manage_lang)%("<i>%s</i>"%attr_id)
        
        # Move down.
        # ----------
        elif key == 'attr' and btn == 'move_down':
          attr_id = REQUEST['attr_id']
          self._moveMetaObjAttr(id,attr_id,+1)
          message = self.getLangStr('MSG_MOVEDOBJDOWN',manage_lang)%("<i>%s</i>"%attr_id)
        
        ##### SYNCHRONIZE ####      
        self.synchronizeObjAttrs()
        
        # Return with message.
        message = urllib.quote(message)
        return RESPONSE.redirect('manage_customizeMetaobjForm?lang=%s&manage_lang=%s&manage_tabs_message=%s&id=%s&attr_id=%s'%(lang,manage_lang,message,id,REQUEST.get('attr_id','')))

###################################################################################################
